<?php
//============================================================+
// File name   : example_001.php
// Begin       : 2008-03-04
// Last Update : 2013-05-14
//
// Description : Example 001 for TCPDF class
//               Default Header and Footer
//
// Author: Nicola Asuni
//
// (c) Copyright:
//               Nicola Asuni
//               Tecnick.com LTD
//               www.tecnick.com
//               info@tecnick.com
//============================================================+

/**
 * Creates an example PDF TEST document using TCPDF
 * @package com.tecnick.tcpdf
 * @abstract TCPDF - Example: Default Header and Footer
 * @author Nicola Asuni
 * @since 2008-03-04
 */

// Include the main TCPDF library (search for installation path).
require_once('tcpdf_include.php');
class MYPDF extends TCPDF {
    public $myheader;
    
    //Page header
    public function Header() {
        // Logo
        switch($this->myheader){
            case 1:
                $this->Cell(0, 15, '<< TCPDF Example '.$this->page.' >>', 0, false, '', 0, '', 0, false, 'M', 'M');
                //$this->Cell(0, 0, '<< TCPDF Example >>', 0, false, 'C', 0, '', 0, false, 'M', 'M');
                break;
        }
    }
}
// create new PDF document
$pdf = new MYPDF(PDF_PAGE_ORIENTATION, PDF_UNIT, PDF_PAGE_FORMAT, true, 'UTF-8', false);
$pdf->myheader=0;

// set document information
$pdf->SetCreator(PDF_CREATOR);
$pdf->SetAuthor('Nicola Asuni');
$pdf->SetTitle('TCPDF Example 001');
$pdf->SetSubject('TCPDF Tutorial');
$pdf->SetKeywords('TCPDF, PDF, example, test, guide');

// set default header data
//$pdf->SetHeaderData(PDF_HEADER_LOGO, PDF_HEADER_LOGO_WIDTH, PDF_HEADER_TITLE, PDF_HEADER_STRING);

// set default header data
$pdf->setFooterData(array(0,64,0), array(0,64,128));

// set header and footer fonts
$pdf->setHeaderFont(Array(PDF_FONT_NAME_MAIN, '', PDF_FONT_SIZE_MAIN));
$pdf->setFooterFont(Array(PDF_FONT_NAME_DATA, '', PDF_FONT_SIZE_DATA));

// set default monospaced font
$pdf->SetDefaultMonospacedFont(PDF_FONT_MONOSPACED);

// set margins
$pdf->SetMargins(PDF_MARGIN_LEFT, 20, PDF_MARGIN_RIGHT);
$pdf->SetHeaderMargin(17);
$pdf->SetFooterMargin(PDF_MARGIN_FOOTER);

// set auto page breaks
$pdf->SetAutoPageBreak(TRUE, PDF_MARGIN_BOTTOM);

// set image scale factor
$pdf->setImageScale(PDF_IMAGE_SCALE_RATIO);

// set some language-dependent strings (optional)
if (@file_exists(dirname(__FILE__).'/lang/eng.php')) {
	require_once(dirname(__FILE__).'/lang/eng.php');
	$pdf->setLanguageArray($l);
}

// ---------------------------------------------------------

// set default font subsetting mode
$pdf->setFontSubsetting(true);

// Set font
// dejavusans is a UTF-8 Unicode font, if you only need to
// print standard ASCII chars, you can use core fonts like
// helvetica or times to reduce file size.

//$pdf->SetFont('dejavusans', '', 14, '', true);

// Add a page
// This method has several options, check the source code documentation for more information.
$pdf->AddPage();

// set text shadow effect
$pdf->setTextShadow(array('enabled'=>true, 'depth_w'=>0.2, 'depth_h'=>0.2, 'color'=>array(196,196,196), 'opacity'=>1, 'blend_mode'=>'Normal'));


// Set some content to print
$html = <<<EOD
<h1>Welcome to <a href="http://www.tcpdf.org" style="text-decoration:none;background-color:#CC0000;color:black;">&nbsp;<span style="color:black;">TC</span><span style="color:white;">PDF</span>&nbsp;</a>!</h1>
<i>This is the first example of TCPDF library.</i>
<p>This text is printed using the <i>writeHTMLCell()</i> method but you can also use: <i>Multicell(), writeHTML(), Write(), Cell() and Text()</i>.</p>
<p>Please check the source code documentation and other examples for further information.</p>
<p style="color:#CC0000;">TO IMPROVE AND EXPAND TCPDF I NEED YOUR SUPPORT, PLEASE <a href="http://sourceforge.net/donate/index.php?group_id=128076">MAKE A DONATION!</a></p>
EOD;
// Print text using writeHTMLCell()
$page_no=$pdf->getPage();
$pdf->writeHTMLCell(0, 0, '', '', $html, 0, 1, 0, true, '', true);

$pdf->cell(0,0, "left".$pdf->getMargins()['left'],0,true);
$pdf->cell(0,0, "right".$pdf->getMargins()['right'],0,1);
$pdf->cell(0,0, "top".$pdf->getMargins()['top'],0,true);
$pdf->cell(0,0, "bottom".$pdf->getMargins()['bottom'],0,true);
$pdf->cell(0,0, "header".$pdf->getMargins()['header'],0,true);
$pdf->cell(0,0, "footer".$pdf->getMargins()['footer'],0,true);
$pdf->myheader=1;
$txt='Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj'.'Nächste Zeile  öadjfakl jföakljfaökljf kaölj fökladjf öalkjdfaökljfaköljfkaöljf ölaj föaklj föal jkf öajf aölkj fkaöljfölafjöflajfölafjaöfjöfajaföaf klaöf flöjakfölfa jfajafj';
//$pdf->( $w, $h = 0, $txt = '', $border = 0, $ln = 0, $align = '', $fill = false, $link = '', $stretch = 0, $ignore_min_height = false, $calign = 'T', $valign = 'M' ) ;
//$pdf->Cell(0,0,$txt ,1,1,'',false, '',0, true, 'T','M');

//MultiCell( $w, $h, $txt, $border = 0, $align = 'J', $fill = false, $ln = 1, $x = '', $y = '', $reseth = true, $stretch = 0, $ishtml = false, $autopadding = true, $maxh = 0, $valign = 'T', $fitcell = false ) 
$pdf->SetY(250);
$pdf->MultiCell(0,0,$txt ,1,'J',false,1);
if($page_no<>$pdf->getPage()){
  $page_no<>$pdf->getPage();
}else{
  $y=140;
  if($pdf->GetY()<$y){
    $pdf->SetY($y);
  }
}

$txt=$pdf->getPage().' Mein Eintrag, gety()'.$pdf->GetY();
//$pdf->( $w, $h = 0, $txt = '', $border = 0, $ln = 0, $align = '', $fill = false, $link = '', $stretch = 0, $ignore_min_height = false, $calign = 'T', $valign = 'M' ) ;
$pdf->Cell(0,0,$txt ,1,1,'',false, '',0, false, 'T','M');

// Close and output PDF document
// This method has several options, check the source code documentation for more information.
$pdf->Output('example_001.pdf', 'I');

//============================================================+
// END OF FILE
//============================================================+
